// Make C_LITERAL nodes.  A literal is Datrix is any kind of *Cst in gcc.
// Datrix literals have their type as an attribute (for some reason)
// rather than 
// (or could be if it were not const), which means:
//	local variables in methods and functions
//	member variables in classes
//	formal parameters (of types, functions, and handlers)
//	bit fields
// It excludes enumerated constants, of course.
// This program only handles concrete C_OBJECT nodes, the subclasses
// (formal parameters) are handled elsewhere.

/* AJM 2001 Mar 14 */

#include <stdio.h>
#include <string.h>
#include <gd_graph.h>
#include "gd_utils.h"

int main (int argc, char *argv [])
{
    void do_literals (gd_graph *);
    char *infilename;
    char *outfilename;
    int i;
    gd_graph g;		// the graph

    if (argc != 3)
    {
      fprintf (stderr, "usage: %s ingdfile outgdfile\n", argv [0]);
      exit (1);
    }

    infilename = argv[1];
    outfilename = argv[2];

    g = gd_graph_read_binary_file (infilename, 1.1, 1.0, 1.1);
    if (g. curStringPos == 0) {
	fprintf (stderr,"can't read gd graph on %s\n", infilename);
	exit (1);
    }

    do_literals (&g);

    if (gd_graph_write_binary_file (outfilename, g) == 0)
    {
	fprintf (stderr, "can't write %s\n", outfilename);
	exit (1);
    }

    return 0;
}

// The following structure contains information about builtin types and
// about formatting their literals.

static struct
{
    char *type_name;		// name of the type as a string
    str_ptr type_str_ptr;	// str_ptr of the type's name, to be looked up
    short width;		// number of bits in the type
    long long int mask;		// mask out bits of given long value
    char *fmt;			// printf format for literals of the type
} formats [] = {
	{ "char",			0, 8,  0x00000000000000FF, "'%s'" },	// format the literal map string
	{ "signed char",		0, 8,  0x00000000000000FF, "'%s'" },
	{ "unsigned char",		0, 8,  0x00000000000000FF, "'%s'" },
	{ "short int",			0, 16, 0x000000000000FFFF, "%hd" },
	{ "short unsigned int",		0, 16, 0x000000000000FFFF, "%huU" },
	{ "float",			0, 32, 0x00000000FFFFFFFF, "%s" },
	{ "int",			0, 32, 0x00000000FFFFFFFF, "%ld" },
	{ "long int",			0, 32, 0x00000000FFFFFFFF, "%ld" },
	{ "unsigned int",		0, 32, 0x00000000FFFFFFFF, "%luU" },
	{ "long unsigned int",		0, 32, 0x00000000FFFFFFFF, "%luU" },
	{ "double",			0, 64, 0xFFFFFFFFFFFFFFFF, "%s" },
	{ "long long int",		0, 64, 0xFFFFFFFFFFFFFFFF, "%lldLL" },
	{ "long long unsigned int",	0, 64, 0xFFFFFFFFFFFFFFFF, "%lluLLU" },
	{ "bool",			0, 1,  0x0000000000000001, "%s"},	// format "true" or "false"
	0
};

// This is a table mapping characters to strings in a string literal.
static char *char_map [] =
{
	"\\x00", "\\x01", "\\x02", "\\x03", "\\x04", "\\x05", "\\x06", "\\a",
	"\\b",   "\\t",   "\\n",   "\\v",   "\\f",   "\\r",   "\\x0E", "\\x0F",
	"\\x10", "\\x11", "\\x12", "\\x13", "\\x14", "\\x15", "\\x16", "\\x17",
	"\\x18", "\\x19", "\\x1a", "\\x1b", "\\x1c", "\\x1d", "\\x1e", "\\x1f",
	" ",     "!",     "\\\"",  "#",     "$",     "%",     "&",     "\\'",
	"(",     ")",     "*",     "+",     ",",     "-",     ".",     "/",
	"0",     "1",     "2",     "3",     "4",     "5",     "6",     "7",
	"8",     "9",     ":",     ";",     "<",     "=",     ">",     "?",
	"@",     "A",     "B",     "C",     "D",     "E",     "F",     "G",
	"H",     "I",     "J",     "K",     "L",     "M",     "N",     "O",
	"P",     "Q",     "R",     "S",     "T",     "U",     "V",     "W",
	"X",     "Y",     "Z",     "[",     "\\\\",  "]",     "^",     "_",
	"`",     "a",     "b",     "c",     "d",     "e",     "f",     "g",
	"h",     "i",     "j",     "k",     "l",     "m",     "n",     "o",
	"p",     "q",     "r",     "s",     "t",     "u",     "v",     "w",
	"x",     "y",     "z",     "{",     "|",     "}",     "~",     "\\x7f",
	"\\x80", "\\x81", "\\x82", "\\x83", "\\x84", "\\x85", "\\x86", "\\x87",
	"\\x88", "\\x89", "\\x8a", "\\x8b", "\\x8c", "\\x8d", "\\x8e", "\\x8f",
	"\\x90", "\\x91", "\\x92", "\\x93", "\\x94", "\\x95", "\\x96", "\\x97",
	"\\x98", "\\x99", "\\x9a", "\\x9b", "\\x9c", "\\x9d", "\\x9e", "\\x9f",
	"\\xa0", "\\xa1", "\\xa2", "\\xa3", "\\xa4", "\\xa5", "\\xa6", "\\xa7",
	"\\xa8", "\\xa9", "\\xaa", "\\xab", "\\xac", "\\xad", "\\xae", "\\xaf",
	"\\xb0", "\\xb1", "\\xb2", "\\xb3", "\\xb4", "\\xb5", "\\xb6", "\\xb7",
	"\\xb8", "\\xb9", "\\xba", "\\xbb", "\\xbc", "\\xbd", "\\xbe", "\\xbf",
	"\\xc0", "\\xc1", "\\xc2", "\\xc3", "\\xc4", "\\xc5", "\\xc6", "\\xc7",
	"\\xc8", "\\xc9", "\\xca", "\\xcb", "\\xcc", "\\xcd", "\\xce", "\\xcf",
	"\\xd0", "\\xd1", "\\xd2", "\\xd3", "\\xd4", "\\xd5", "\\xd6", "\\xd7",
	"\\xd8", "\\xd9", "\\xda", "\\xdb", "\\xdc", "\\xdd", "\\xde", "\\xdf",
	"\\xe0", "\\xe1", "\\xe2", "\\xe3", "\\xe4", "\\xe5", "\\xe6", "\\xe7",
	"\\xe8", "\\xe9", "\\xea", "\\xeb", "\\xec", "\\xed", "\\xee", "\\xef",
	"\\xf0", "\\xf1", "\\xf2", "\\xf3", "\\xf4", "\\xf5", "\\xf6", "\\xf7",
	"\\xf8", "\\xf9", "\\xfa", "\\xfb", "\\xfc", "\\xfd", "\\xfe", "\\xff"
};

static int char_map_len [256];

// This function formats the indicated string as a string literal, by
// adding initial and final double-quotes, and by escaping what needs to be escaped.

str_ptr string_literal (gd_graph *g, str_ptr p)
{
    unsigned char *s;	// original string
    char len;	// its length

    unsigned char *c;

    char *s_lit;	// literal string
    int lit_len;	// its length

    str_ptr result;

    s = & g-> stringBuffer [p];
    len = strlen (s);

    // compute length of literal
    lit_len = 2;		// for the double quotes
    for (c = s; *c; ++c)
	lit_len = lit_len + char_map_len [*c];
    
    s_lit = (char *) malloc (lit_len+1);	// 1 for the final null.
    *s_lit = '\0';				// initially emptyu

    strcat (s_lit, "\"");			// initial doublequote
    for (c = s; *c; ++c)
	strcat (s_lit, char_map [*c]);		// each escape.
    strcat (s_lit, "\"");			// final doublequote

    result = gd_add_string (g, s_lit);
    free (s_lit);

    return result;
}

void do_literals (gd_graph *g)
{
    int i;
    node_ptr ptr;			// Ptr to node which is to be converted to object.

    gd_graph_node new_literal;		// New Datrix node for object.
    gd_graph_node new_edge;		// New Datrix edge for various purposes.

    long long val;			// Supplied 64-bit literal's value.

    unsigned char bit1;			// The above as masked down to shorter values.
    unsigned char bit8;
    unsigned short bit16;
    unsigned long bit32;
    unsigned long long bit64;		// This is gcc C!

    char literal [50];
    str_ptr literal_ptr;		// Ptr to the literal string after formatting.

    int msgn = 0;

    // Compute lengths of char map.
    for (i = 0; i < 256; ++i)
	char_map_len [i] = strlen (char_map [i]);

    // Lookup the builtin type names.
    for (i = 0; formats [i]. type_name != 0; ++i)
	formats [i]. type_str_ptr = gd_lookup_string (g, formats [i]. type_name);

    // Do all the literals.
    for (ptr = 0; ptr < g-> curNodePos; ++ ptr)
    {
	// Various kinds of constant.
	literal_ptr = -1;		// set to ptr if successful

	if (g-> nodeBuffer [ptr]. common. type == gSTRING_CST)
	    literal_ptr = string_literal (g, g-> nodeBuffer [ptr]. string_cst. strg);

	else
	if (g-> nodeBuffer [ptr]. common. type == gREAL_CST) /* T.D. */
	    literal_ptr = string_literal (g, g-> nodeBuffer [ptr]. real_cst. value);
	else
	if (g-> nodeBuffer [ptr]. common. type == gREAL_CST || g-> nodeBuffer [ptr]. common. type == gINTEGER_CST)
	{
	    // This is a builtin-type numeric literal, but its format depends on type.

	    // Construct its complete supplied value (of type long long int).
	    // Currently the value is lacking for REAL_CST values.
	    if (g-> nodeBuffer [ptr]. common. type == gINTEGER_CST)
		val =
			((long long unsigned) (g-> nodeBuffer [ptr]. integer_cst. high)) << 32 
		    |	((long long unsigned) (g-> nodeBuffer [ptr]. integer_cst. low));
	    else
	    	val = 0;

//	    fprintf (stderr, "[%d] literal value 0x%016llx of type %s (%d)",
//		    ptr,
//		    val,
//		    &(g-> stringBuffer [g-> nodeBuffer [g-> nodeBuffer [ptr]. gcc_cst. type]. dc_identifier. name]),
//		    g-> nodeBuffer [g-> nodeBuffer [ptr]. gcc_cst. type]. dc_identifier. name
//		);

	    // Search for the formatting types.
	    for (i = 0; formats [i]. type_name != 0; ++i)
		if (formats [i]. type_str_ptr == g-> nodeBuffer [g-> nodeBuffer [ptr]. gcc_cst. type]. dc_identifier. name)
		{
		    switch (formats [i]. width)
		    {
		    case 1:
			bit1 = (unsigned char) (formats [i]. mask & val);
			sprintf (literal, formats [i]. fmt, bit1? "true": "false");
			break;
		    case 8:
			bit8 = (unsigned char) (formats [i]. mask & val);
			sprintf (literal, formats [i]. fmt, char_map [bit8]);
			break;
		    case 16:
			bit16 = (unsigned short) (formats [i]. mask & val);
			sprintf (literal, formats [i]. fmt, bit16);
			break;
		    case 32:
			bit32 = (unsigned long) (formats [i]. mask & val);
			sprintf (literal, formats [i]. fmt, bit32);
			break;
		    case 64:
			bit64 = (unsigned long long) (formats [i]. mask & val);
			sprintf (literal, formats [i]. fmt, bit64);
			break;
		    }
		    break;
		}
	    if (formats [i]. type_name == 0)
		// Couldn't find it.  Assume int.
		// (Proper for enumerated constants, btw.)
		sprintf (literal, "%d", (int) (0xffffffff & val));

	    literal_ptr = gd_add_string (g, literal);
	}
	
	if (literal_ptr != -1)
	{
	    new_literal. common. type = dC_LITERAL;
	    new_literal. common. other_node = -1;
	    new_literal. dc_expression. refers = -1;
	    new_literal. dc_literal. value = literal_ptr;
	    new_literal. dc_literal. type = newNode (g);

	    new_edge. common. type = dC_INSTANCE;
	    new_edge. common. other_node = -1;
	    new_edge. dc_arc. ordinal = 0;
	    new_edge. dc_arc. src = ptr;
	    new_edge. dc_arc. dest = g-> nodeBuffer [ptr]. gcc_cst. type;
	    new_edge. dc_arc. next = -1;
	    g-> nodeBuffer [new_literal. dc_literal. type] = new_edge;

	    g-> nodeBuffer [ptr] = new_literal;
	}
    }
}
