// In the Datrix model, every dc_identifier is supposed to be syntactically
// embedded in a unique place: and therefore all edges except one should
// pt at the identifier indirectly, that is, by means of a dc_name_ref
// indirection node and a dc_refers_to edge.

// This transform creates all those dc_name_ref nodes and dc_refers_to edges
// for each Datrix edge which is not the embedding edge.

/* AJM 2001 Mar 29 */

#include <stdio.h>
#include <string.h>
#include <gd_graph.h>
#include "gd_utils.h"

int main (int argc, char *argv [])
{
    void do_namerefs (gd_graph *);
    char *infilename;
    char *outfilename;
    int i;
    gd_graph g;		// the graph

    if (argc != 3)
    {
      fprintf (stderr, "usage: %s ingdfile outgdfile\n", argv [0]);
      exit (1);
    }

    infilename = argv[1];
    outfilename = argv[2];

    g = gd_graph_read_binary_file (infilename, 1.1, 1.0, 1.3);
    if (g. curStringPos == 0) {
	fprintf (stderr,"can't read gd graph on %s\n", infilename);
	exit (1);
    }

    do_namerefs (&g);

    if (gd_graph_write_binary_file (outfilename, g) == 0)
    {
	fprintf (stderr, "can't write %s\n", outfilename);
	exit (1);
    }

    return 0;
}

// This function sets the mark_flag1 bit for any edge and edge destination on the
// next-edge chain starting at p, when the edge destination is not an expression.

void set_edge_flag1 (gd_graph *g, node_ptr p)
{
    node_ptr cur;

    for (
	cur = p;
	    0 <= cur && cur < g-> curNodePos
	&&  g-> nodeBuffer [cur]. common. type >= dC_ARC
	&&  g-> nodeBuffer [cur]. common. type <= dC_ARC_ARRAY_DIM;
	cur = g-> nodeBuffer [cur]. dc_arc. next
	)
    {
	// Check it's a valid dest.
	node_ptr dest = g-> nodeBuffer [cur]. dc_arc. dest;
	if (0 <= dest && dest <= g-> curNodePos)
	{
	    if (
		    dC_EXPRESSION <= g-> nodeBuffer [dest]. common. type
		&&  g-> nodeBuffer [dest]. common. type <= dC_DESTROY_EXPR
		)
		continue;

	    g-> nodeBuffer [cur]. common. mark_flag1 = 1;
	    g-> nodeBuffer [dest]. common. mark_flag1 = 1;
	}
    }
}

void do_namerefs (gd_graph *g)
{
    int i;
    gd_graph_node *p;			// Pointer to node which is to be converted to object.
    node_ptr ptr;			// Ptr to node which is to be converted to object.

    gd_graph_node new_node;		// New Datrix node for the goto statement
    gd_graph_node new_edge;		// New Datrix edge for the arcson edge

    // We'll do this by first tagging every edge (using the common.mark_flag1)
    // as the embedding edge (1) or not (0).  Then traverse the actual edges
    // doing the nameref construction.

    // Clear the flags of everything.
    for (ptr = 0; ptr < g-> curNodePos; ++ ptr)
	g-> nodeBuffer [ptr]. common. mark_flag1 = 0;
    
    // Set the flags by walking structures where declarations can be.
    // Mark embedding edges and also embedded things.
    // With the exception of for initializers, because until we know what's
    // embedded elsewhere, we can't tell if a for loop initializer is a declaration
    // there, or not.

    for (ptr = 0; ptr < g-> curNodePos; ++ ptr)
    {
	switch (g-> nodeBuffer [ptr]. common. type)
	{
	case dC_AGGR_TYPE:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_aggr_type. members);
	    break;
	case dC_ENUM_TYPE:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_enum_type. values);
	    break;
	case dC_GENERIC_TYPE:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_generic_type. formals);
	    break;
	case dC_FCT_TYPE:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_fct_type. formalParams);
	    break;
	case dC_FUNCTION:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_function. formal_fct);
	    break;
	case dC_SCOPE:
	case dC_SYSTEM:
	case dC_SCOPE_GLB:
	case dC_SCOPE_COMPIL:
	case dC_SCOPE_FILE:
	case dC_PACKAGE:
	case dC_NAME_SPACE:
	case dC_BLOCK:
	case dC_STATIC_BLOCK:
	case dC_TRY_BLOCK:
	case dC_CATCH_BLOCK:
	case dC_FINALLY_BLOCK:
	case dC_SYNCHRONIZED_BLOCK:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_scope. contents);
	    break;
	case dC_FOR_LOOP:
	    set_edge_flag1 (g, g-> nodeBuffer [ptr]. dc_for_loop. init);
	    break;
	default:
	    break;
	}
    }

    // Now pass through the whole thing again, looking at nonembedding edges.
    for (ptr = 0; ptr < g-> curNodePos; ++ ptr)
	if (
		g-> nodeBuffer [ptr]. common. mark_flag1 == 0
	    &&  dC_ARC <= g-> nodeBuffer [ptr]. common. type
	    &&  g-> nodeBuffer [ptr]. common. type <= dC_ARC_ARRAY_DIM
	    &&  g-> nodeBuffer [ptr]. common. type != dC_REFERS_TO
	    )
	{
	    node_ptr dest = g-> nodeBuffer [ptr]. dc_arc. dest;
	    enum gd_node_type type = g-> nodeBuffer [dest]. common. type;

	    // Defensively check dest is valid.
	    if (0 <= dest && dest < g-> curNodePos); else continue;

	    // Check that this is a declarable type of node.
	    // ... is an identifier
	    if (dC_IDENTIFIER <= type && type < dC_EXPRESSION); else continue;
	    // ... is not a scope
	    if (dC_SCOPE <= type && type < dC_OBJECT) continue;
	    // ... is not a type (why?)
	    if (dC_TYPE <= type && type < dC_SCOPE) continue;

	    // This is a not-embedding structural edge to a named identifier edge.
	    // Make a new node for the name ref and a new edge for the refers to.
	    new_node. common. type = dC_NAME_REF;
	    new_node. common. other_node = -1;
	    new_node. dc_expression. refers = newNode (g);	// for new refers-to edge.
	    new_node. dc_name_ref. name = g-> nodeBuffer [dest]. dc_identifier. name;

	    new_edge. common. type = dC_REFERS_TO;
	    new_edge. common. other_node = -1;
	    new_edge. dc_arc. ordinal = 0;
	    new_edge. dc_arc. src = newNode (g);		// for new name_ref node
	    new_edge. dc_arc. dest = g-> nodeBuffer [ptr]. dc_arc. dest;
 	    new_edge. dc_arc. next = -1;

	    g-> nodeBuffer [ptr]. dc_arc. dest = new_edge. dc_arc. src;
	    g-> nodeBuffer [new_node. dc_expression. refers] = new_edge;
	    g-> nodeBuffer [new_edge. dc_arc. src] = new_node;
	}
}
