/*+
 * Name:	merge_builtin.c
 *
 * Purpose:	GCC may generate more than one instance of a builtin type.
 * 		This program merges multiple instance of each builtin type
 * 		into a single node.
 *
 * Author:	Thomas R. Dean
 *
 * Date:	May 4, 2001
 *
-*/

#include <stdio.h>
#include <stdlib.h>

#include "gd_graph.h"
#include "gd_utils.h"

// Defined in gd_walkers.c
void walk_ptrs (gd_graph *, void (*) (gd_graph *, node_ptr, node_ptr *, int));

// Defined below...
void mergeBuiltin (gd_graph *, node_ptr, node_ptr *, int);


char * progname;

void usage(){
  fprintf(stderr,"usage: %s infile outfile\n",progname);
  exit(1);
}

#define num_global_builtins  100
node_ptr global_builtins[num_global_builtins];
int cur_global_builtin = 0;
node_ptr lastArc = -1;
node_ptr origLastArc = -1;

int main(int argc, char * argv[]){
  gd_graph g;
  char * filename, *outfilename;
  node_ptr i;
  node_ptr rover;
  progname = argv[0];

  if (argc != 3) {
    usage();
  }

  filename = argv[1];
  outfilename = argv[2];

  g = gd_graph_read_binary_file(filename,1.0,1.0,1.0);
  if (g.curStringPos == 0) {
    fprintf(stderr,"could not get graph\n");
    exit(1);
  }

  /* start at front node */
  for(i = 0; i < g.curNodePos; i++){
    g.nodeBuffer[i].common.mark_flag1 = 0;
  }

  if (g.nodeBuffer[g.firstNode].common.type != dC_NAME_SPACE){
    fprintf(stderr,"first node is not a namespace node!!\n");
    exit(1);
  }

  rover = g.nodeBuffer[g.firstNode].dc_scope.contents;
  if (rover != -1){
    if (g.nodeBuffer[rover].common.type != dC_ARC_SON){
      fprintf(stderr,"contents of namespace not arcs!!\n");
      exit(1);
    }
  }

  /* mark built in types that are sons of global namespace*/
  while(rover != -1) {
    node_ptr dest = g.nodeBuffer[rover].dc_arc.dest;
    if (g.nodeBuffer[dest].common.type == dC_BUILT_IN_TYPE){
      /* mark as son of global namespace */
      g.nodeBuffer[i].common.mark_flag1 = 1;
      /* remember the type */
      if(cur_global_builtin == num_global_builtins) {
	fprintf(stderr,"more than %d global built in types?",num_global_builtins);
	exit(1);
      }
      global_builtins[cur_global_builtin++] = dest;
    }
    /* go to the next arc */
    if (g.nodeBuffer[rover].dc_arc.next == -1){
      lastArc = rover;
      origLastArc = rover;
    }
    rover = g.nodeBuffer[rover].dc_arc.next;
  }
  /* we now have a list of the built in types that are direct
   * sons of the global namespace. Now we traverse all working pointers
   * and point any duplicates at the "official" version. If we encounter
   * a new built in type that is not in the list, we add it, after making
   * that node a son of the global namespace */
  walk_ptrs(&g,mergeBuiltin);

  /* verify only one ArcSon edge for moved types */
  /* AFAIK all built in types that are not sons of global name
   * space in GCC are floating, not son of anything. But we
   * check anyways */
  if(lastArc != origLastArc) {
    rover = origLastArc;
    while(rover != -1){
      for(i = 0; i < g.curNodePos; i++){
	if (i != rover){
	  if (g.nodeBuffer[i].common.type == dC_ARC_SON) {
	    if (g.nodeBuffer[i].dc_arc.dest ==
		g.nodeBuffer[rover].dc_arc.dest){
	      fprintf(stderr,"node %d has more than one parent",rover);
	      exit(1);
	    }
	  }
	}
      }
    }
  }
 

  for(i = 0; i < g.curNodePos; i++){
    g.nodeBuffer[i].common.mark_flag1 = 0;
  }

  /* all done, write the file */
  if (gd_graph_write_binary_file(outfilename,g) == 0){
    fprintf(stderr,"problem writing file\n");
    exit(1);
  }

  return 0;
}

void mergeBuiltin (gd_graph * g, node_ptr p, node_ptr * ref_ptr, int n) {
  /* skip other_node */
  if (n == 0)
    return;

  // Ignore nonptrs.
  if (*ref_ptr < -1 && *ref_ptr >= g-> curNodePos)
    return;

  if (g->nodeBuffer[*ref_ptr].common.type == dC_BUILT_IN_TYPE){
    if (g->nodeBuffer[*ref_ptr].common.mark_flag1 == 0){
      /* this built in type is not a son of the global namespace */
      int i = 0;
      while(i < cur_global_builtin) {
	if (g->nodeBuffer[global_builtins[i]].dc_identifier.name ==
	    g->nodeBuffer[*ref_ptr].dc_identifier.name){
	  /* found a match in the list of "official" types */
	  *ref_ptr = global_builtins[i];
	  break;
	}
	i++;
      }
      if (i == cur_global_builtin){
	/* we did not find it in the list of "official" types */
	/* add as son to global namespace and mark as official */
	node_ptr newArc = newNode(g);
	g->nodeBuffer[newArc].common.type = dC_ARC_SON;
	g->nodeBuffer[newArc].common.other_node = -1;
	g->nodeBuffer[newArc].dc_arc.src = g->firstNode;
	g->nodeBuffer[newArc].dc_arc.dest= *ref_ptr;
	g->nodeBuffer[newArc].dc_arc.ordinal = 0;
	g->nodeBuffer[newArc].dc_arc.next = -1;
	/* add to global namespace arc list */
	g->nodeBuffer[lastArc].dc_arc.next = newArc;
	/* add to list of global built in types */
	if(cur_global_builtin == num_global_builtins) {
	  fprintf(stderr,"more than %d global built in types?",num_global_builtins);
	  exit(1);
	}
	g->nodeBuffer[*ref_ptr].common.mark_flag1 = 1;
	global_builtins[cur_global_builtin++] = *ref_ptr;
      }
    }
  }
}
