/*+
 * Name:	ordinals.c
 *
 * Purpose:	Fixes ordinals attributes in all datrix arcs
 *
 * Author:	Thomas R. Dean
 *
 * Date:	April 10, 2001
 *
-*/

#include <stdio.h>
#include <stdlib.h>

#include "gd_graph.h"
#include "gd_utils.h"

char * progname;

void usage(){
  fprintf(stderr,"usage: %s infile outfile\n",progname);
  exit(1);
}

int main(int argc, char * argv[]){
  gd_graph g;
  char * filename, *outfilename;
  int nodenum,i, count;
  node_ptr rover;
  str_ptr void_ptr;
  progname = argv[0];

  if (argc != 3) {
    usage();
  }

  filename = argv[1];
  outfilename = argv[2];

  g = gd_graph_read_binary_file(filename,1.0,1.0,1.0);
  if (g.curStringPos == 0) {
    fprintf(stderr,"could not get graph\n");
    exit(1);
  }

  /* first make sure the mark flag is clear */
  for(i = 0; i < g.curNodePos; i++){
    g.nodeBuffer[i].common.mark_flag1 = 0;
  }

  /* for all arcs, clear the ordinal and if
   * the next pointer is valid, set the mark bit on the
   * next edge */
  for(i = 0; i < g.curNodePos; i++){
    if (g.nodeBuffer[i].common.type >= dC_ARC &&
	g.nodeBuffer[i].common.type <= dC_ARC_ARRAY_DIM){
      g.nodeBuffer[i].dc_arc.ordinal = 0;
      if (g.nodeBuffer[i].dc_arc.next < -1 ||
	  g.nodeBuffer[i].dc_arc.next >= (int) g.curNodePos){
	fprintf(stderr,"Next Field of arc %d is out of range %d\n",i,
	    g.nodeBuffer[i].dc_arc.next);
	exit(1);
      } else {
	if (g.nodeBuffer[i].dc_arc.next != -1){
	  g.nodeBuffer[g.nodeBuffer[i].dc_arc.next].common.mark_flag1 = 1;
	}
      }
    }
  }

  /* at this point in time, all arc nodes with the mark bit
   * clear are the beginning of arc chains, if there is
   * an arc with the bit clear and has a valid next pointer,
   * number the arcs */

  for(i = 0; i < g.curNodePos; i++){
    if (g.nodeBuffer[i].common.type >= dC_ARC &&
	g.nodeBuffer[i].common.type <= dC_ARC_ARRAY_DIM){
      if (g.nodeBuffer[i].common.mark_flag1 == 0 && 
	  g.nodeBuffer[i].dc_arc.next != -1){
	/* both unmarked AND more than one arc in the chain */
	/* generate ordinal attributes */
	int ctr = 1;
	rover = i;
	while (rover != -1){
	  /* assert that all nodes are arcs */
	  if (g.nodeBuffer[rover].common.type < dC_ARC ||
	      g.nodeBuffer[rover].common.type > dC_ARC_ARRAY_DIM){
	    fprintf(stderr, "non arc node (%d) in arc chain?\n",rover);
	    exit(1);
	  }

	  g.nodeBuffer[rover].dc_arc.ordinal = ctr++;
	  rover = g.nodeBuffer[rover].dc_arc.next;
	}
      }
    }
  }

  /* special fix for Function nodes. Function nodes
   * have two sets of ArcSon edges, one for parameters,
   * one for the body of the function. Make the body
   * list continue from the parameter list */
  /* if the function does not have any parameters,
   * then the only arcson chain is the body and it will
   * be numbered correctly */

  for(i = 0; i < g.curNodePos; i++){
    if (g.nodeBuffer[i].common.type == dC_FUNCTION){
      rover = g.nodeBuffer[i].dc_function.formal_fct;
      if (rover != -1){
	int ctr = 1;
	/* if there was only one parameter, the arcsons will
	 * not be numbered. We have to traverse to the end
	 * of them anywas, so it costs little more to renumber them
	 * all anyways. */
	while (rover != -1){
	  g.nodeBuffer[rover].dc_arc.ordinal = ctr++;
	  rover = g.nodeBuffer[rover].dc_arc.next;
	}
	/* ctr is now appropriate for the rest of the function */
	rover = g.nodeBuffer[i].dc_function.exception;
	while (rover != -1){
	  g.nodeBuffer[rover].dc_arc.ordinal = ctr++;
	  rover = g.nodeBuffer[rover].dc_arc.next;
	}
	rover = g.nodeBuffer[i].dc_function.block;
	while (rover != -1){
	  g.nodeBuffer[rover].dc_arc.ordinal = ctr++;
	  rover = g.nodeBuffer[rover].dc_arc.next;
	}
      }
    }
  }

  /* cleanup */
  for(i = 0; i < g.curNodePos; i++){
    g.nodeBuffer[i].common.mark_flag1 = 0;
  }

  /* all done, write the file */
  if (gd_graph_write_binary_file(outfilename,g) == 0){
    fprintf(stderr,"problem writing file\n");
    exit(1);
  }

  return 0;
}

